/*
 * draw.cpp
 *
 * Copyright (c) Stephen Thompson, 2007.
 * Licenced for non-commercial use only. See LICENCE.txt for details.
 *
 * The various in-game drawing routines
 *
 */

#include "misc.hpp"

#include "cfg.hpp"
#include "draw.hpp"
#include "platform/display.hpp"
#include "platform/graphic.hpp"


//
// drawBackpackEntry
//

void DrawUI::drawBackpackEntry(Display &display, int left, int bottom, int right, int top,
							   int slot, const Graphic *gfx, int no_carried, int no_max)
{
	if (!gfx) return;
	if (no_carried == 0) return;

	display.clearClipRectangle();
	
	// Slots: 101, 102, ... 199 are Normal Items
	// 200 = Lock Picks
	// 201, 202, 203 = Keys.
	// (Anything else is an error, and is ignored.)

	if (slot > 100 && slot < 200) {
		// Normal items are drawn in a column format.
		// Vertical spacing is chosen such that no_max items fit exactly into the height
		// of the inventory area. (If no_max==0, then space them by 4 pixels -- perhaps
		// this should be configurable though?)

		const int x = left + GetWidth(*gfx)*(slot-101);
		const int ht = GetHeight(*gfx);

		if (no_max != 0) {
			const int num = top - bottom - ht;
			if (no_carried > no_max) no_max = no_carried;
			for (int j = no_max - no_carried + 1; j <= no_max; ++j) {
				display.drawGraphic(x, num*j/no_max, *gfx, false);
			}
		} else {
			// spacing of 4; extra ones are just dropped.
			const int jmax = (top - ht - bottom) / 4;
			for (int j = min(no_carried-1, jmax); j>=0; --j) {
				display.drawGraphic(x, top - ht - j*4, *gfx, false);
			}
		}

	} else {
		// Keys are drawn downwards from the top, except that the lock picks
		// (slot 200) are always at the bottom.
		// NB no_carried and no_max are ignored for keys and lockpicks.
		const int x = right - GetWidth(*gfx);
		const int ht = GetHeight(*gfx);
		const int y = slot==200 ? top - ht*4 : top - ht*(slot-200);
		display.drawGraphic(x, y, *gfx, false);
	}
}	


//
// drawMessage
//

void DrawUI::drawMessage(Display &display, int bl_x, int bl_y, int tr_x, int tr_y,
						 int sq_rel_x, int sq_rel_y, int w, int h, const string &message)
{
	const int pixels_per_square = CfgInt("pixels_per_square");
	const int room_bl_x = (tr_x - bl_x - w*pixels_per_square)/2 + bl_x;
	const int room_bl_y = (tr_y - bl_y - h*pixels_per_square)/2 + bl_y;
	const int sq_bl_x = sq_rel_x*pixels_per_square + room_bl_x;
	const int sq_bl_y = sq_rel_y*pixels_per_square + room_bl_y;
	const int text_width = display.getTextWidth(message.c_str());
	const int tx = sq_bl_x + (pixels_per_square/2) - (text_width/2);
	const int ty = sq_bl_y + pixels_per_square*11/10;
	display.setClipRectangle(bl_x, bl_y, tr_x, tr_y);	
	display.drawText(tx, ty, message.c_str());
}


//
// drawMenu
// NOTE: we assume that all graphics are the same size.
//

void DrawUI::drawMenu(Display &display, int bl_x, int bl_y,
					  const Graphic * gfx_centre, const Graphic * gfx_empty,
					  const Graphic * gfx_highlight,
					  const Graphic * gfx_north, const Graphic * gfx_east,
					  const Graphic * gfx_south, const Graphic * gfx_west,
					  bool highlight, MapDirection highlight_dir)
{
	display.clearClipRectangle();
	if (gfx_centre) {
		const int w = GetWidth(*gfx_centre);
		const int h = GetHeight(*gfx_centre);
		display.drawGraphic(bl_x + w, bl_y + h, *gfx_centre);
		drawMenuItem(display, gfx_north, gfx_empty, bl_x + w, bl_y + 2*h);
		drawMenuItem(display, gfx_west, gfx_empty, bl_x, bl_y + h);
		drawMenuItem(display, gfx_east, gfx_empty, bl_x + 2*w, bl_y + h);
		drawMenuItem(display, gfx_south, gfx_empty, bl_x + w, bl_y);
		if (gfx_highlight) {
			int xofs = 0, yofs = 0;
			if (highlight) {
				switch (highlight_dir) {
				case D_NORTH:
					yofs = h;
					break;
				case D_EAST:
					xofs = w;
					break;
				case D_SOUTH:
					yofs = -h;
					break;
				case D_WEST:
					xofs = -w;
					break;
				}
			}
			display.drawGraphic(bl_x + xofs + w, bl_y + yofs + h, *gfx_highlight);
		}
	}
}

void DrawUI::drawMenuItem(Display &disp, const Graphic *g, const Graphic *e, int x, int y)
{
	if (g) disp.drawGraphic(x,y,*g);
	else if (e) disp.drawGraphic(x,y,*e);
}



