/*
 * entity.cpp
 *
 * Copyright (c) Stephen Thompson, 2007.
 * Licenced for non-commercial use only. See LICENCE.txt for details.
 *
 */

#include "misc.hpp"

#include "anim.hpp"
#include "cfg.hpp"
#include "entity.hpp"
#include "mediator.hpp"
#include "task_manager.hpp"

//
// function that computes travel times
//

namespace {
	int TravelTime(int dist, int speed, int points_per_square)
	{
		return CfgInt("walk_time") * dist * 100 / (speed * points_per_square);
	}
}

//
// MotionTask: helper class. Stops motion when an entity reaches its destination.
//

class MotionTask : public Task {
public:
	MotionTask(weak_ptr<Entity> a) : agt(a) { }
	virtual void execute(TaskManager &);
private:
	weak_ptr<Entity> agt;
};


void MotionTask::execute(TaskManager &tm)
{
	// Check entity is valid
	shared_ptr<Entity> entity(agt.lock());
	if (!entity) return;
	if (!entity->getMap()) return;

	// the entity may have stopped moving and/or started a different
	// move with a different arrival time, so check for this.
	// (The motion task should always run at arrival_time - 1.)
	if (!entity->isMoving()) return;
	if (tm.getGVT() < entity->getArrivalTime()-1) return;

	// Stop the motion
	MotionType mt = entity->getMotionType();
	if (mt == MT_MOVE) {
		// moved a complete square. shift position forward one square.
		entity->reposition(DisplaceCoord(entity->getPos(), entity->getFacing()));
	} else {
		entity->flags &= (~Entity::MT_BITS);
		if (mt == MT_WITHDRAW) entity->flags &= (~Entity::APPROACHING_BIT);
	}
	entity->arrival_time = 0;

	// Tell mediator
	// nb. entity may have been removed from map by onReposition events; we must not call
	// Mediator::onChangeEntityMotion if this is the case.
	if (entity->getMap()) {
		Mediator::instance().onChangeEntityMotion(entity);
	}
}



//
// Entity
//

Entity::Entity()
	: dmap(0), anim(0), anim_frame(0), anim_tzero(0), anim_invuln(false), overlay(0),
	  start_time(0), arrival_time(0), speed(100), flags(VISIBLE_BIT)
{ }


//
// getNearestPos
//

MapCoord Entity::getNearestPos() const
{
	static const int halfway_point = CfgInt("points_per_square") / 2;
	if (getOffset() < halfway_point) return getPos();
	else return DisplaceCoord(getPos(), getFacing());
}


//
// map positioning
//

void Entity::addToMap(DungeonMap *new_dmap, const MapCoord &new_pos)
{
	if (dmap || !new_dmap) return;
	
	// Set motion data.
	cancelAllMotion();

	// Add to the map
	dmap = new_dmap;
	pos = new_pos;
	MapHelper::addEntity(shared_from_this());

	// Tell mediator
	Mediator::instance().onAddEntity(shared_from_this());
}

void Entity::rmFromMap()
{
	if (!dmap) return;
	
	// Tell mediator
	Mediator::instance().onRmEntity(shared_from_this());

	// Remove from the map
	MapHelper::rmEntity(shared_from_this());
	dmap = 0;
}

void Entity::reposition(const MapCoord &new_pos)
{
	if (!dmap) return;
	shared_ptr<Entity> self(shared_from_this());

	// Remove from map
	MapHelper::rmEntity(self);
	
	// Set new position, and cancel motion
	pos = new_pos;
	cancelAllMotion();

	// Add to map
	MapHelper::addEntity(self);

	// Tell mediator
	Mediator::instance().onRepositionEntity(shared_from_this());
}

void Entity::setFacing(MapDirection dir)
{
	if (isMoving() || isApproaching()) return;

	// Change the facing direction
	flags &= (~FACING_MASK);
	flags |= (int(dir) << FACING_SHIFT);

	// Tell mediator
	Mediator::instance().onChangeEntityFacing(shared_from_this());
}



//
// visibility
//

void Entity::setVisible(bool v)
{
	if (isVisible() != v) {
		if (v) flags |= VISIBLE_BIT;
		else flags &= (~VISIBLE_BIT);
		Mediator::instance().onChangeEntityVisible(shared_from_this());
	}
}


//
// Anim/overlay routines
//

void Entity::setAnim(const Anim * new_anim)
{
	if (anim != new_anim) {
		anim = new_anim;
		if (getMap()) {
			// setAnim can be called from the ctor, before shared_from_this is available
			// Hence the check "if (getMap())" above.
			Mediator::instance().onChangeEntityAnim(shared_from_this());
		}
	}
}

void Entity::setAnimFrame(int frame, int tzero)
{
	if (anim_frame != frame || anim_tzero != tzero) {
		anim_frame = frame;
		anim_tzero = tzero;
		Mediator::instance().onChangeEntityAnim(shared_from_this());
	}
}

void Entity::setAnimInvulnerability(bool i)
{
	if (anim_invuln != i) {
		anim_invuln = i;
		Mediator::instance().onChangeEntityAnim(shared_from_this());
	}
}

void Entity::setOverlay(const Overlay *new_overlay)
{
	if (overlay != new_overlay) {
		overlay = new_overlay;
		if (getMap()) { // prevent problems if overlay changed before shared_from_this() is available
			Mediator::instance().onChangeEntityAnim(shared_from_this());
		}
	}
}

int Entity::getAnimFrame() const
{
	if (dmap) return (Mediator::instance().getGVT() < anim_tzero || anim_tzero <= 0)
				  ? anim_frame : 0;
	else return 0;
}

void Entity::getAnimData(const Anim *&a, const Overlay *&ov, int &f, int &zt, bool &inv) const
{
	a = anim;
	ov = overlay;
	f = getAnimFrame();
	zt = anim_tzero;
	inv = anim_invuln;
}



//
// Motion routines (see also MotionTask)
//

void Entity::move(MotionType mt)
{
	static const int approach_offset = CfgInt("approach_offset");
	static const int points_per_square = CfgInt("points_per_square");

	// Get some constants
	if (!dmap) return;
	if (mt == MT_NOT_MOVING) return;
	if (getSpeed() <= 0) return;
	TaskManager &tm(Mediator::instance().getTaskManager());
	const int gvt = tm.getGVT();
	shared_ptr<Entity> self(shared_from_this());
	
	// Remove from map
	MapHelper::rmEntity(self);

	// Compute distance
	int dist;
	if (mt == MT_APPROACH || mt == MT_WITHDRAW) {
		dist = approach_offset;
	} else {
		ASSERT(mt == MT_MOVE);
		dist = points_per_square;
		if (isApproaching()) dist -= approach_offset;
	}

	// Set motion flags and arrival_time
	const int travel_time = TravelTime(dist, getSpeed(), points_per_square);
	start_time = gvt;
	arrival_time = travel_time + gvt;
	flags = (flags & ~(MT_BITS+APPROACHING_BIT)) | int(mt);
	if (mt == MT_APPROACH || mt == MT_WITHDRAW) flags |= APPROACHING_BIT;

	// Add to map
	MapHelper::addEntity(self);

	// Set a MotionTask, to be run when the motion finishes.
	shared_ptr<MotionTask> task(new MotionTask(self));
	tm.addTask(task, TP_NORMAL, arrival_time - 1);

	// Tell Mediator.
	Mediator::instance().onChangeEntityMotion(self);
}

void Entity::flipMotion()
{
	static const int points_per_square = CfgInt("points_per_square");

	if (!dmap) return;
	if (!isMoving() || isApproaching()) return;
	if (getSpeed() <= 0) return;
	TaskManager &tm(Mediator::instance().getTaskManager());
	const int gvt = tm.getGVT();
	shared_ptr<Entity> self(shared_from_this());

	// Compute the new motion parameters
	int old_ofs = getOffset();
	if (old_ofs == 0) old_ofs = 1;
	const int travel_time = TravelTime(old_ofs, getSpeed(), points_per_square);
	const int new_arrival_time = gvt + travel_time;
	const int new_start_time = new_arrival_time - points_per_square * travel_time / old_ofs;

	// Update the entity
	MapHelper::rmEntity(self);
	const MapDirection old_facing = getFacing();
	pos = DisplaceCoord(pos, old_facing);
	flags &= (~FACING_MASK);
	flags |= (int(Opposite(old_facing)) << FACING_SHIFT);
	start_time = new_start_time;
	arrival_time = new_arrival_time;
	MapHelper::addEntity(self);

	// Add a new MotionTask
	// (The old one will automatically cancel itself when it executes.)
	shared_ptr<MotionTask> task(new MotionTask(self));
	tm.addTask(task, TP_NORMAL, arrival_time - 1);

	// Tell Mediator.
	Mediator::instance().onFlipEntityMotion(self);
}


int Entity::getOffset() const
{
	static const int approach_offset = CfgInt("approach_offset");
	static const int points_per_square = CfgInt("points_per_square");
	
	if (!getMap()) return 0;
	const MotionType mt = getMotionType();
	if (mt == MT_NOT_MOVING) {
		if (isApproaching()) return approach_offset;
		else return 0;
	} else {
		int sp, ep;
		switch (mt) {
		case MT_MOVE:
			sp = 0;
			ep = points_per_square;
			break;
		case MT_APPROACH:
			sp = 0;
			ep = approach_offset;
			break;
		case MT_WITHDRAW:
			sp = approach_offset;
			ep = 0;
			break;
		default:
			ASSERT(0);
			break;
		}
		const int gvt = Mediator::instance().getGVT();
		return sp + (ep - sp) * (gvt - start_time) / (arrival_time - start_time);
	}
}

