/*
 * knights_game.cpp
 *
 * Copyright (c) Stephen Thompson, 2007.
 * Licenced for non-commercial use only. See LICENCE.txt for details.
 *
 */

#include "misc.hpp"

#include "cfg.hpp"
#include "control_actions.hpp"
#include "keyboard_controller.hpp"
#include "knights_game.hpp"
#include "menu_selections.hpp"
#include "platform/display.hpp"
#include "platform/sound_manager.hpp"
#include "rstream/rstream.hpp"
#include "stylus_controller.hpp"
#include "title_screen.hpp"

#ifdef _WIN32_WCE
#include "app/gapi_app2d.hpp"
#else
#include "app/sdl_app2d.hpp"
#endif

#include "boost/scoped_ptr.hpp"
using namespace boost;

/////////////////////////////////////////////////////////////
// this controls whether to open windowed or fullscreen:
/////////////////////////////////////////////////////////////

const bool windowed = true;



/////////////////////////////////////////////////////
// One-Time Initialization
/////////////////////////////////////////////////////

namespace {
	scoped_ptr<App2D> app2d;
	scoped_ptr<Display> display;
}

void Initialize(const string &resource_dir)
{
#ifdef _WIN32_WCE
	const wchar_t * windowname = L"Knights";
	const wchar_t * classname = L"Knights Window Class";
#endif
	const char * game_name = "Knights";

	// initialize resource lib
	RStream::Initialize(resource_dir);

	// initialize "Cfg" library
	CfgInit();

	// Create app and open window

	// Which app to create will depend on the platform:
#ifdef _WIN32_WCE
	app2d.reset(new GapiApp2D(classname, windowname));
#else
	app2d.reset(new SdlApp2D);
#endif

#ifdef _WIN32_WCE
	// WinCE: simplified logic since we always open full-screen in mode zero. 
	app2d->openFullScreen(0, game_name);
#else
	// Other platforms: We can use either full-screen or windowed mode
	bool open = false;
	if (!windowed) {
		int nmodes = app2d->findDisplayModes();
		for (int i=0; i<nmodes; ++i) {
			DisplayMode dm = app2d->getDisplayMode(i);
			if (dm.width == 800 && dm.height == 600 && dm.colour_depth >= 16) {
				app2d->openFullScreen(i, game_name);
				open = true;
				break;
			}
		}
	}
	if (!open) {
		app2d->openWindowed(0,0,660,462,game_name);
	}
#endif		
	
	// Set up the standard controls
	SetupStandardControls();
	
	// Setup Display and Controllers.
	// (TODO) In future controllers should probably be loaded from some sort
	// of config file.

	display.reset(new Display);

	SoundManager::instance();  // init sound at the beginning, rather than on-demand.
}



///////////////////////////////////////////////////////////
// Quit Key for WinCE
// (since there is no other way to close the program!)
///////////////////////////////////////////////////////////

#ifdef _WIN32_WCE

bool quit_key_enabled = true;
void EnableQuitKey() {
	quit_key_enabled = true;
}
void DisableQuitKey() {
	quit_key_enabled = false;
}
class QuitKey : public AppObserver {
public:
	void key(Keycode c)
	{
		if (c == KC_LEFT && quit_key_enabled) {
			app2d->exitMainLoop();
		}
	}
};
#endif



//////////////////////////////////////////////
// Main Routine (RunKnights)
//////////////////////////////////////////////

void RunKnights(const string &resource_dir)
{
	UsefulStuff us;

	// Construct the app, display etc.
	Initialize(resource_dir);

	// Setup controllers
	{
#ifdef _WIN32_WCE
		StylusController left_ctrlr(Keycode('A'), Keycode('B'), Keycode('S'),
									*app2d, display->getHeight());
		StylusController right_ctrlr(Keycode('A'), Keycode('B'), Keycode('S'),
									 *app2d, display->getHeight());
#else
		KeyboardController left_ctrlr(Keycode('W'), Keycode('D'), Keycode('S'),
									  Keycode('A'), Keycode('Q'), KC_LEFT_ALT, *app2d);
		KeyboardController right_ctrlr(KC_UP, KC_RIGHT, KC_DOWN, KC_LEFT,
									   KC_RIGHT_CONTROL, KC_RIGHT_ALT, *app2d);
#endif
		
		// Setup UsefulStuff
		us.app2d = app2d.get();
		us.display = display.get();
		us.left_controller = &left_ctrlr;
		us.right_controller = &right_ctrlr;

#ifdef _WIN32_WCE
		// Add the QuitKey observer
		QuitKey qk;
		app2d->addObserver(&qk);
#endif
		
		// Start up TitleScreen.
		SetupTitleScreen(us);
		app2d->enterMainLoop();

#ifdef _WIN32_WCE
		// remove the QuitKey observer when done
		app2d->rmObserver(&qk);
#endif
	}

	// Delete things
	// The controllers will have been deleted already by now.
	us.remote.reset();
	us.server_config.reset();
	us.client_config.reset();
	SoundManager::destroyInstance();  // ensure sound is shut down before App...
	display.reset();
	app2d.reset();
}

