/*
 * colour_change.hpp
 *
 * Copyright (c) Stephen Thompson, 2007.
 * Licenced for non-commercial use only. See LICENCE.txt for details.
 *
 * Defines a mapping from old colours to new colours. Used for
 * dynamically changing colours of Graphics, eg for Knight House
 * Colours, or for changing the colour of your potion bottle when you
 * have strength, super etc.
 *
 * NOTE: Colours are quantized to 5-bit accuracy in each of the R, G
 * and B components. This is to accommodate different display formats
 * (including 16-bit "555" format). A proper solution would probably
 * be to specify what "accuracy" you want when constructing the
 * ColourChange object (eg for a 16-bit display you would use 5-bit
 * accuracy; for a 24-bit display, 8-bit, etc). However, 5-bit
 * accuracy works fine for everything I want to use this for...
 * 
 */

#ifndef COLOUR_CHANGE_HPP
#define COLOUR_CHANGE_HPP

#include <vector>
using namespace std;

// The RGB struct is useful in other contexts so it is made public.
struct RGB {
	int r, g, b;
	bool operator<(const RGB &other) const {
		return (r<<16) + (g<<8) + b	< (other.r << 16) + (other.g << 8) + other.b;
	};
	bool operator==(const RGB &other) const {
		return r==other.r && g==other.g && b==other.b;
	}
};


class ColourChange {
public:
	// Constructor: creates an 'empty' ColourChange object (where all colours are passed
	// through unmodified).
	ColourChange() { }
	
	// empty(): true if this is the 'identity' colour change
	bool empty() const { return mappings.empty(); }
	
	// add: adds a mapping from (ro, go, bo) to (rn, gn, bn) [where
	// o=old, n=new] where r,g,b are values in the range 0..255.
	void add(int ro, int go, int bo, int rn, int gn, int bn);

	// lookup: returns false if the src colour is to be left unchanged
	// returns true (and sets rn,gn,bn) if the src colour is to be
	// changed to (rn, gn, bn).
	bool lookup(int ro, int go, int bo, int &rn, int &gn, int &bn) const;

	// comparison functions.
	bool operator<(const ColourChange &other) const { return mappings < other.mappings; }
	bool operator==(const ColourChange &other) const { return mappings == other.mappings; }

	// Serialization interface
	// Note we don't bother with ID's for colour changes. Instead, the same ColourChange
	// will be written several times if necessary.
	void serialize(ostream &) const;
	explicit ColourChange(istream &);
	
private:
	// Accuracy mask -- this is what enforces the 5-bit comparison accuracy.
	enum { MASK = 0xf8 };
	
	// Vector of (old,new) pairs. This is kept sorted (on "old" values) at all times.
	vector<pair<RGB,RGB> > mappings;
};

#endif
